const throttled = function (handler, timeout) {
  if (!timeout) return handler;
  let active = false;
  let event = null;
  return function (e) {
    event = e;
    if (active) return;
    active = true;
    setTimeout(() => {
      handler(event);
      active = false;
    }, timeout);
  };
};

const event = (event, {
  target,
  throttle
} = {}) => {
  const funcName = `_${event}Handler`;
  const deactivatedName = `_${event}Deactivated`;
  if (!target) target = window;
  return {
    mounted() {
      this[funcName] = throttled(this[event + 'Handler'], throttle);
      target.addEventListener(event, this[funcName]);
    },

    destroyed() {
      target.removeEventListener(event, this[funcName]);
    },

    activated() {
      if (!this[deactivatedName]) return;
      target.addEventListener(event, this[funcName]);
      this[deactivatedName] = false;
    },

    deactivated() {
      target.removeEventListener(event, this[funcName]);
      this[deactivatedName] = true;
    }

  };
};
const scroll = event('scroll', {
  throttle: 17
});
const resize = event('resize', {
  throttle: 17
});

const eps = 0.05;

const isNumber = (...values) => values.every(value => typeof value == 'number');

const pxEq = (first, second) => {
  return isNumber(first, second) && Math.abs(first - second) < eps;
};
const pxGt = (first, second) => {
  return isNumber(first, second) && first - second > eps;
};

//
var script = {
  mixins: [scroll, resize],
  props: {
    // Array of data that is passed to item
    itemsData: {
      type: Array,
      required: true
    },
    // Approximate item height (used at first render)
    itemHeight: {
      type: Number,
      required: true
    },
    // Vue js item component
    itemComponent: {
      type: [Object, Function],
      required: true
    },
    // Height of rendered part relative to viewport height
    renderViewports: {
      type: Number,
      default: 3
    },
    // The name of the item component function that is called on item resize
    onItemResize: String
  },

  data() {
    return {
      start: 0,
      // index of the first rendered item
      end: 0,
      // index of the last rendered item (not included)
      height: 0,
      // average item height
      spacerMargin: 0 // margin before spacer element

    };
  },

  computed: {
    visibleItems() {
      return this.itemsData.slice(this.start, this.end).map((v, i) => {
        return {
          data: v,
          index: this.start + i
        };
      });
    },

    spacerStyle() {
      return {
        transform: 'translateY(' + this.spacerMargin + 'px)'
      };
    },

    listStyle() {
      return {
        height: this.height + 'px',
        position: 'relative',
        overflow: 'hidden',
        overflowAnchor: 'none' // onItemsResize does this

      };
    }

  },
  methods: {
    /**
     * Scroll event handler
     * Checks scrolled distance. If the distance is within rendered items,
     * it will change start and spacerMargin according to known (or average)
     * items sizes.
     * If the scrolled distance is bigger, it will change start
     * and spacerMargin to approximate values using average item height
     */
    scrollHandler() {
      const prevScrollTop = this.scrollTop;
      this.scrollTop = Math.max(0, -this.$refs.list.getBoundingClientRect().top);
      const diff = this.scrollTop - prevScrollTop;
      const preRendered = Math.ceil(Math.ceil(window.innerHeight / this.average) * (this.renderViewports - 1) / 2);

      const height = index => this.heights.get(index) || this.average;

      const heights = (index, down = false) => {
        let res = 0;
        const max = index + preRendered + (down ? 1 : 0);

        for (let i = index; i < max; i++) res += height(i);

        return res;
      };

      if (Math.abs(diff) < 0.65 * window.innerHeight * this.renderViewports) {
        let {
          spacerMargin,
          start
        } = this;

        if (diff > 0) {
          while (spacerMargin + heights(start, true) < this.scrollTop) {
            spacerMargin += height(start);
            start++;
          }
        } else {
          while (start > 0 && spacerMargin + heights(start) > this.scrollTop) {
            start--;
            spacerMargin -= height(start);
          }
        }

        this.spacerMargin = spacerMargin;
        this.setStart(start);
      } else {
        const start = Math.floor(this.scrollTop / this.average);
        this.spacerMargin = start * this.average;
        this.setStart(start);
      }
    },

    /**
     * Resize event handler
     * Clears known items' heights if window width changes
     * Keeps visible heights
     */
    resizeHandler() {
      if (!pxEq(this.prevWidth, this.width())) {
        const old = this.heights;
        this.heights = new Map();

        for (let i = this.start; i < this.end; i++) if (old.has(i)) this.heights.set(i, old.get(i));
      }
    },

    /**
     * Getting real item height from dom
     * Uses the distance between bottoms of current and previous items.
     * If it's the first rendered item, uses top of the list element
     * It allows to use items with margins
     */
    getItemHeight(index) {
      const {
        spacer
      } = this.$refs;
      index -= this.start;
      const prev = index === 0 ? this.$refs.list.getBoundingClientRect().top + this.spacerMargin : spacer.children[index - 1].getBoundingClientRect().bottom;
      return spacer.children[index].getBoundingClientRect().bottom - prev;
    },

    /**
     * Returns true if item's top is invisible
     * Used to check if it needs to correct spacerMargin when actual size
     * of items changes.
     */
    isTopItem(index) {
      const {
        spacer
      } = this.$refs;
      return spacer.children[index - this.start].getBoundingClientRect().top > 0;
    },

    /**
     * ResizeObserver handler
     * Updates average items' height and corrects spacerMargin
     */
    onItemsResize(entities) {
      if (this.isDeactivated) return;
      let {
        average
      } = this;
      average *= this.heights.size;
      let changed = false;
      let spacerMargin = 0;
      entities.forEach(entity => {
        const el = entity.target;
        const idx = Number(el.dataset.itemIndex);

        if (idx >= this.start && idx < this.end) {
          const prevHeight = this.heights.get(idx);
          const newHeight = this.getItemHeight(idx);

          if (!pxEq(prevHeight, newHeight)) {
            if (!this.isTopItem(idx)) spacerMargin += newHeight - (prevHeight || this.average);
            average += newHeight - (prevHeight || 0);
            this.heights.set(idx, newHeight);
            changed = true;
          }

          if (this.onItemResize) el.__vue__[this.onItemResize]();
        }
      });

      if (changed) {
        this.average = average / this.heights.size;
        this.spacerMargin -= spacerMargin;
        this.setStart(this.start);
      }

      this.$nextTick(() => {
        this.fixSpacerMargin();
        this.fixHeight();
      });
    },

    observe(from, to) {
      this.$nextTick(() => {
        for (let i = from; i < to; i++) this.observer.observe(this.$refs.spacer.children[i]);
      });
    },

    unobserve(from, to) {
      for (let i = from; i < to; i++) this.observer.unobserve(this.$refs.spacer.children[i]);
    },

    /**
     * Sets a new start index
     * Calculates end index and updates observed items in ResizeObserver
     */
    setStart(index) {
      this.prevWidth = this.width();
      const count = Math.round(this.renderViewports * Math.ceil(window.innerHeight / this.average));
      const {
        start: prevStart,
        end: prevEnd
      } = this;
      this.start = Math.max(index, 0);
      this.end = Math.min(this.start + count, this.itemsData.length);
      if (prevStart === this.start && prevEnd === this.end) return;
      if (this.start != prevStart || this.end != prevEnd) this.unobserve(0, Math.min(this.start, prevEnd) - prevStart);
      if (prevEnd > this.end) this.unobserve(Math.max(this.end, prevStart) - prevStart, prevEnd - prevStart);
      if (this.start < prevStart) this.observe(0, Math.min(prevStart, this.end) - this.start);
      if (prevEnd < this.end) this.observe(Math.max(prevEnd, this.start) - this.start, this.end - this.start);
      const bottom = this.itemsData.length === this.end;
      if (bottom && !this.bottom) this.$emit('bottom');
      this.bottom = bottom;
    },

    /**
     * Fixes spacerMargin errors
     */
    fixSpacerMargin() {
      if (this.start === 0) {
        if (pxEq(this.spacerMargin, 0)) return;
        window.scroll(0, window.scrollY - this.spacerMargin);
        this.spacerMargin = 0;
      } else if (pxGt(0, this.spacerMargin)) {
        const avMargin = this.spacerMargin - this.average * this.start;
        this.spacerMargin -= avMargin;
        window.scroll(0, window.scrollY - avMargin);
      }
    },

    /**
     * Fixes height error
     */
    fixHeight() {
      const {
        list,
        spacer
      } = this.$refs;
      const diff = spacer.getBoundingClientRect().bottom - list.getBoundingClientRect().bottom;

      if (this.itemsData.length == this.end) {
        if (pxEq(diff, 0)) return;
        this.height += diff;
      } else if (pxGt(diff, 0)) {
        this.height = Math.max(this.height + diff, this.average * this.itemsData.length);
      } else if (Math.abs(this.height - this.average * this.itemsData.length) > 0.1 * this.itemsData.length * this.average) {
        this.height = this.average * this.itemsData.length;
      }
    },

    width() {
      const {
        list
      } = this.$refs;
      return list ? list.getBoundingClientRect().width : 0;
    }

  },
  watch: {
    itemsData() {
      this.setStart(this.start);
    }

  },

  created() {
    this.heights = new Map();
    this.scrollTop = 0;
    this.average = this.itemHeight;
    this.observer = new ResizeObserver(this.onItemsResize);
    this.setStart(0);
  },

  beforeDestroy() {
    this.observer.disconnect();
  },

  activated() {
    this.isDeactivated = false;
  },

  deactivated() {
    this.isDeactivated = true;
  }

};

function normalizeComponent(template, style, script, scopeId, isFunctionalTemplate, moduleIdentifier /* server only */, shadowMode, createInjector, createInjectorSSR, createInjectorShadow) {
    if (typeof shadowMode !== 'boolean') {
        createInjectorSSR = createInjector;
        createInjector = shadowMode;
        shadowMode = false;
    }
    // Vue.extend constructor export interop.
    const options = typeof script === 'function' ? script.options : script;
    // render functions
    if (template && template.render) {
        options.render = template.render;
        options.staticRenderFns = template.staticRenderFns;
        options._compiled = true;
        // functional template
        if (isFunctionalTemplate) {
            options.functional = true;
        }
    }
    // scopedId
    if (scopeId) {
        options._scopeId = scopeId;
    }
    let hook;
    if (moduleIdentifier) {
        // server build
        hook = function (context) {
            // 2.3 injection
            context =
                context || // cached call
                    (this.$vnode && this.$vnode.ssrContext) || // stateful
                    (this.parent && this.parent.$vnode && this.parent.$vnode.ssrContext); // functional
            // 2.2 with runInNewContext: true
            if (!context && typeof __VUE_SSR_CONTEXT__ !== 'undefined') {
                context = __VUE_SSR_CONTEXT__;
            }
            // inject component styles
            if (style) {
                style.call(this, createInjectorSSR(context));
            }
            // register component module identifier for async chunk inference
            if (context && context._registeredComponents) {
                context._registeredComponents.add(moduleIdentifier);
            }
        };
        // used by ssr in case component is cached and beforeCreate
        // never gets called
        options._ssrRegister = hook;
    }
    else if (style) {
        hook = shadowMode
            ? function (context) {
                style.call(this, createInjectorShadow(context, this.$root.$options.shadowRoot));
            }
            : function (context) {
                style.call(this, createInjector(context));
            };
    }
    if (hook) {
        if (options.functional) {
            // register for functional component in vue file
            const originalRender = options.render;
            options.render = function renderWithStyleInjection(h, context) {
                hook.call(context);
                return originalRender(h, context);
            };
        }
        else {
            // inject component registration as beforeCreate hook
            const existing = options.beforeCreate;
            options.beforeCreate = existing ? [].concat(existing, hook) : [hook];
        }
    }
    return script;
}

/* script */
const __vue_script__ = script;
/* template */

var __vue_render__ = function () {
  var _vm = this;

  var _h = _vm.$createElement;

  var _c = _vm._self._c || _h;

  return _c('div', {
    ref: "list",
    style: _vm.listStyle
  }, [_c('div', {
    ref: "spacer",
    style: _vm.spacerStyle
  }, _vm._l(_vm.visibleItems, function (item) {
    return _c(_vm.itemComponent, {
      key: item.index,
      tag: "component",
      attrs: {
        "data-item-index": item.index,
        "data": item.data,
        "index": item.index
      }
    });
  }), 1)]);
};

var __vue_staticRenderFns__ = [];
/* style */

const __vue_inject_styles__ = undefined;
/* scoped */

const __vue_scope_id__ = undefined;
/* module identifier */

const __vue_module_identifier__ = undefined;
/* functional template */

const __vue_is_functional_template__ = false;
/* style inject */

/* style inject SSR */

/* style inject shadow dom */

const __vue_component__ = /*#__PURE__*/normalizeComponent({
  render: __vue_render__,
  staticRenderFns: __vue_staticRenderFns__
}, __vue_inject_styles__, __vue_script__, __vue_scope_id__, __vue_is_functional_template__, __vue_module_identifier__, false, undefined, undefined, undefined);

// Import vue component

const install = function installListScroller(Vue) {
  if (install.installed) return;
  install.installed = true;
  Vue.component('ListScroller', __vue_component__);
}; // Create module definition for Vue.use()
// to be registered via Vue.use() as well as Vue.component()


__vue_component__.install = install; // Export component by default
// also be used as directives, etc. - eg. import { RollupDemoDirective } from 'rollup-demo';
// export const RollupDemoDirective = component;

export default __vue_component__;
